@@ -50,6 +50,8 @@
 #include <osmocom/abis/ipa.h>
 #include <osmocom/core/backtrace.h>
 #include <osmocom/gsm/ipa.h>
+#include <osmocom/core/stats_tcp.h>
+#include <osmocom/core/fsm.h>
 
 /* global parameters of IPA input driver */
 struct ipa_pars g_e1inp_ipaccess_pars;
@@ -76,7 +78,7 @@ static inline void ipaccess_keepalive_fsm_cleanup(struct e1inp_ts *e1i_ts)
 
 	ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
 	if (ka_fsm) {
-		ipa_keepalive_fsm_stop(ka_fsm);
+		osmo_fsm_inst_term(ka_fsm, OSMO_FSM_TERM_REQUEST, NULL);
 		e1i_ts->driver.ipaccess.ka_fsm = NULL;
 	}
 }
@@ -122,7 +124,8 @@ static int ipaccess_drop(struct osmo_fd *bfd, struct e1inp_line *line)
 	e1i_ts->pending_msg = NULL;
 
 	/* e1inp_sign_link_destroy releases the socket descriptors for us. */
-	line->ops->sign_link_down(line);
+	if (line->ops->sign_link_down)
+		line->ops->sign_link_down(line);
 
 	e1inp_line_put2(line, __func__);
 	return ret;
@@ -295,6 +298,7 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 			struct e1inp_ts *ts;
 			struct osmo_fd *newbfd;
 			struct e1inp_line *new_line;
+			char tcp_stat_name[64];
 
 			sign_link =
 				line->ops->sign_link_up(&unit_data, line,
@@ -341,6 +345,9 @@ static int ipaccess_rcvmsg(struct e1inp_line *line, struct msgb *msg,
 				     "could not register FD\n");
 				goto err;
 			}
+			snprintf(tcp_stat_name, sizeof(tcp_stat_name), "site.%u.bts.%u.ipa-rsl.%u",
+				unit_data.site_id, unit_data.bts_id, unit_data.trx_id);
+			osmo_stats_tcp_osmo_fd_register(newbfd, tcp_stat_name);
 
 			e1i_ts = ipaccess_line_ts(newbfd, new_line);
 			ipaccess_bsc_keepalive_fsm_alloc(e1i_ts, newbfd, "rsl_bsc_to_bts");
@@ -447,15 +454,11 @@ static void ipaccess_close(struct e1inp_sign_link *sign_link)
 	struct e1inp_ts *e1i_ts = sign_link->ts;
 	struct osmo_fd *bfd = &e1i_ts->driver.ipaccess.fd;
 	struct e1inp_line *line = e1i_ts->line;
+	struct osmo_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
 
-	/* line might not exist if != bsc||bts */
-	if (line) {
-		/* depending on caller the fsm might be dead */
-		struct osmo_fsm_inst* ka_fsm = ipaccess_line_ts(bfd, line)->driver.ipaccess.ka_fsm;
-		if (ka_fsm)
-			ipa_keepalive_fsm_stop(ka_fsm);
-
-	}
+	/* depending on caller the fsm might be dead */
+	if (ka_fsm)
+		ipa_keepalive_fsm_stop(ka_fsm);
 
 	e1inp_int_snd_event(e1i_ts, sign_link, S_L_INP_TEI_DN);
 	/* the first e1inp_sign_link_destroy call closes the socket. */
@@ -472,12 +475,24 @@ static void ipaccess_close(struct e1inp_sign_link *sign_link)
 	}
 }
 
+static bool e1i_ts_has_pending_tx_msgs(struct e1inp_ts *e1i_ts)
+{
+	struct e1inp_sign_link *link;
+	llist_for_each_entry(link, &e1i_ts->sign.sign_links, list) {
+		if (!llist_empty(&link->tx_list)) {
+			return true;
+		}
+	}
+	return false;
+}
+
 static void timeout_ts1_write(void *data)
 {
 	struct e1inp_ts *e1i_ts = (struct e1inp_ts *)data;
 
 	/* trigger write of ts1, due to tx delay timer */
-	ts_want_write(e1i_ts);
+	if (e1i_ts_has_pending_tx_msgs(e1i_ts))
+		ts_want_write(e1i_ts);
 }
 
 static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
@@ -489,12 +504,12 @@ static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
 	int ret;
 
 	e1i_ts = ipaccess_line_ts(bfd, line);
-	osmo_fd_write_disable(bfd);
 
 	/* get the next msg for this timeslot */
 	msg = e1inp_tx_ts(e1i_ts, &sign_link);
 	if (!msg) {
 		/* no message after tx delay timer */
+		osmo_fd_write_disable(bfd);
 		return 0;
 	}
 
@@ -504,7 +519,7 @@ static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
 	case E1INP_SIGN_OSMO:
 		break;
 	default:
-		osmo_fd_write_enable(bfd); /* come back for more msg */
+		/* leave WRITE flag enabled, come back for more msg */
 		ret = -EINVAL;
 		goto out;
 	}
@@ -522,13 +537,19 @@ static int __handle_ts1_write(struct osmo_fd *bfd, struct e1inp_line *line)
 		goto err;
 	}
 
-	/* set tx delay timer for next event */
-	osmo_timer_setup(&e1i_ts->sign.tx_timer, timeout_ts1_write, e1i_ts);
-
-	/* Reducing this might break the nanoBTS 900 init. */
-	osmo_timer_schedule(&e1i_ts->sign.tx_timer, 0, e1i_ts->sign.delay);
-
+	/* this is some ancient code that apparently exists to slow down writes towards
+	 * some even more ancient nanoBTS 900 units. See git commit
+	 * d49fc5ae24fc9d44d2b284392ab619cc7a69a876 of openbsc.git (now osmo-bsc.git) */
+	if (e1i_ts->sign.delay) {
+		osmo_fd_write_disable(bfd);
+		/* set tx delay timer for next event */
+		osmo_timer_setup(&e1i_ts->sign.tx_timer, timeout_ts1_write, e1i_ts);
+		osmo_timer_schedule(&e1i_ts->sign.tx_timer, 0, e1i_ts->sign.delay);
+	} else {
 out:
+		if (!e1i_ts_has_pending_tx_msgs(e1i_ts))
+			osmo_fd_write_disable(bfd);
+	}
 	msgb_free(msg);
 	return ret;
 err:
@@ -585,12 +606,11 @@ static void update_fd_settings(struct e1inp_line *line, int fd)
 		val = 1;
 		ret = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &val, sizeof(val));
 		if (ret < 0)
-			LOGP(DLINP, LOGL_NOTICE, "Failed to set keepalive: %s\n",
+			LOGP(DLINP, LOGL_ERROR, "Failed to set keepalive: %s\n",
 			     strerror(errno));
 		else
 			LOGP(DLINP, LOGL_NOTICE, "Keepalive is set: %i\n", ret);
 
-#if defined(TCP_KEEPIDLE) && defined(TCP_KEEPINTVL) && defined(TCP_KEEPCNT)
 		/* The following options are not portable! */
 		val = line->keepalive_idle_timeout > 0 ?
 			line->keepalive_idle_timeout :
@@ -598,7 +618,7 @@ static void update_fd_settings(struct e1inp_line *line, int fd)
 		ret = setsockopt(fd, IPPROTO_TCP, TCP_KEEPIDLE,
 				 &val, sizeof(val));
 		if (ret < 0)
-			LOGP(DLINP, LOGL_NOTICE,
+			LOGP(DLINP, LOGL_ERROR,
 			     "Failed to set keepalive idle time: %s\n",
 			     strerror(errno));
 		val = line->keepalive_probe_interval > -1 ?
@@ -607,7 +627,7 @@ static void update_fd_settings(struct e1inp_line *line, int fd)
 		ret = setsockopt(fd, IPPROTO_TCP, TCP_KEEPINTVL,
 				 &val, sizeof(val));
 		if (ret < 0)
-			LOGP(DLINP, LOGL_NOTICE,
+			LOGP(DLINP, LOGL_ERROR,
 			     "Failed to set keepalive interval: %s\n",
 			     strerror(errno));
 		val = line->keepalive_num_probes > 0 ?
@@ -616,21 +636,18 @@ static void update_fd_settings(struct e1inp_line *line, int fd)
 		ret = setsockopt(fd, IPPROTO_TCP, TCP_KEEPCNT,
 				 &val, sizeof(val));
 		if (ret < 0)
-			LOGP(DLINP, LOGL_NOTICE,
+			LOGP(DLINP, LOGL_ERROR,
 			     "Failed to set keepalive count: %s\n",
 			     strerror(errno));
-#if defined(TCP_USER_TIMEOUT)
                 val = 1000 * line->keepalive_num_probes *
                         line->keepalive_probe_interval +
                         line->keepalive_idle_timeout;
                 ret = setsockopt(fd, IPPROTO_TCP, TCP_USER_TIMEOUT,
                                  &val, sizeof(val));
                 if (ret < 0)
-                        LOGP(DLINP, LOGL_NOTICE,
+			LOGP(DLINP, LOGL_ERROR,
                              "Failed to set user timoeut: %s\n",
                              strerror(errno));
-#endif
-#endif
 	}
 
 	val = 1;
@@ -671,6 +688,7 @@ static int ipaccess_bsc_oml_cb(struct ipa_server_link *link, int fd)
 		LOGP(DLINP, LOGL_ERROR, "could not register FD\n");
 		goto err_line;
 	}
+	osmo_stats_tcp_osmo_fd_register(bfd, "ipa-oml");
 
 	update_fd_settings(line, bfd->fd);
 
@@ -724,6 +742,8 @@ static int ipaccess_bsc_rsl_cb(struct ipa_server_link *link, int fd)
 		LOGP(DLINP, LOGL_ERROR, "could not register FD\n");
 		goto err_line;
 	}
+	osmo_stats_tcp_osmo_fd_register(bfd, "ipa-rsl");
+
 	/* Request ID. FIXME: request LOCATION, HW/SW VErsion, Unit Name, Serno */
 	ret = ipa_ccm_send_id_req(bfd->fd);
 	if (ret < 0) {
@@ -844,9 +864,10 @@ static struct msgb *ipa_bts_id_ack(void)
 static void ipaccess_bts_updown_cb(struct ipa_client_conn *link, int up)
 {
 	struct e1inp_line *line = link->line;
+	struct e1inp_ts *e1i_ts = ipaccess_line_ts(link->ofd, line);
 
 	if (up) {
-		struct osmo_fsm_inst *ka_fsm = ipaccess_line_ts(link->ofd, line)->driver.ipaccess.ka_fsm;
+		struct osmo_fsm_inst *ka_fsm = e1i_ts->driver.ipaccess.ka_fsm;
 
 		update_fd_settings(line, link->ofd->fd);
 		if (ka_fsm && line->ipa_kap)
@@ -854,6 +875,7 @@ static void ipaccess_bts_updown_cb(struct ipa_client_conn *link, int up)
 		return;
 	}
 
+	ipaccess_keepalive_fsm_cleanup(e1i_ts);
 	if (line->ops->sign_link_down)
 		line->ops->sign_link_down(line);
 }
@@ -1071,7 +1093,7 @@ static int ipaccess_line_update(struct e1inp_line *line)
 	}
 	case E1INP_LINE_R_BTS: {
 		struct ipa_client_conn *link;
-		struct e1inp_ts *e1i_ts;
+		struct e1inp_ts *e1i_ts = e1inp_line_ipa_oml_ts(line);
 
 		LOGP(DLINP, LOGL_NOTICE, "enabling ipaccess BTS mode, "
 		     "OML connecting to %s:%u\n", line->ops->cfg.ipa.addr,
@@ -1080,6 +1102,7 @@ static int ipaccess_line_update(struct e1inp_line *line)
 		/* Drop previous line */
 		if (il->ipa_cli[0]) {
 			ipa_client_conn_close(il->ipa_cli[0]);
+			ipaccess_keepalive_fsm_cleanup(e1i_ts);
 			ipa_client_conn_destroy(il->ipa_cli[0]);
 			il->ipa_cli[0] = NULL;
 		}
@@ -1101,7 +1124,7 @@ static int ipaccess_line_update(struct e1inp_line *line)
 		}
 		link->dscp = g_e1inp_ipaccess_pars.oml.dscp;
 		link->priority = g_e1inp_ipaccess_pars.oml.priority;
-		if (ipa_client_conn_open(link) < 0) {
+		if (ipa_client_conn_open2(link, line->connect_timeout) < 0) {
 			LOGP(DLINP, LOGL_ERROR, "cannot open OML BTS link: %s\n",
 				strerror(errno));
 			ipa_client_conn_close(link);
@@ -1109,7 +1132,6 @@ static int ipaccess_line_update(struct e1inp_line *line)
 			return -EIO;
 		}
 
-		e1i_ts = e1inp_line_ipa_oml_ts(line);
 		ipaccess_bts_keepalive_fsm_alloc(e1i_ts, link, "oml_bts_to_bsc");
 		il->ipa_cli[0] = link;
 		ret = 0;
@@ -1123,7 +1145,6 @@ static int ipaccess_line_update(struct e1inp_line *line)
 	return ret;
 }
 
-
 /* backwards compatibility */
 int e1inp_ipa_bts_rsl_connect(struct e1inp_line *line,
 			      const char *rem_addr, uint16_t rem_port)
@@ -1138,6 +1159,7 @@ int e1inp_ipa_bts_rsl_connect_n(struct e1inp_line *line,
 	struct ipa_client_conn *rsl_link;
 	struct e1inp_ts *e1i_ts = e1inp_line_ipa_rsl_ts(line, trx_nr);
 	struct ipaccess_line *il;
+	int rc;
 
 	if (E1INP_SIGN_RSL+trx_nr-1 >= NUM_E1_TS) {
 		LOGP(DLINP, LOGL_ERROR, "cannot create RSL BTS link: "
@@ -1145,17 +1167,14 @@ int e1inp_ipa_bts_rsl_connect_n(struct e1inp_line *line,
 		return -EINVAL;
 	}
 
+	/* Drop previous line */
+	if ((rc = e1inp_ipa_bts_rsl_close_n(line, trx_nr)) < 0)
+		return rc;
+
 	if (!line->driver_data)
 		line->driver_data = talloc_zero(line, struct ipaccess_line);
 	il = line->driver_data;
 
-	/* Drop previous line */
-	if (il->ipa_cli[1 + trx_nr]) {
-		ipa_client_conn_close(il->ipa_cli[1 + trx_nr]);
-		ipa_client_conn_destroy(il->ipa_cli[1 + trx_nr]);
-		il->ipa_cli[1 + trx_nr] = NULL;
-	}
-
 	rsl_link = ipa_client_conn_create2(tall_ipa_ctx,
 					  e1inp_line_ipa_rsl_ts(line, trx_nr),
 					  E1INP_SIGN_RSL+trx_nr,
@@ -1172,7 +1191,7 @@ int e1inp_ipa_bts_rsl_connect_n(struct e1inp_line *line,
 	}
 	rsl_link->dscp = g_e1inp_ipaccess_pars.rsl.dscp;
 	rsl_link->priority = g_e1inp_ipaccess_pars.rsl.priority;
-	if (ipa_client_conn_open(rsl_link) < 0) {
+	if (ipa_client_conn_open2(rsl_link, line->connect_timeout) < 0) {
 		LOGP(DLINP, LOGL_ERROR, "cannot open RSL BTS link: %s\n",
 			strerror(errno));
 		ipa_client_conn_close(rsl_link);
@@ -1184,6 +1203,29 @@ int e1inp_ipa_bts_rsl_connect_n(struct e1inp_line *line,
 	return 0;
 }
 
+/* Close the underlying IPA TCP socket of an RSL link */
+int e1inp_ipa_bts_rsl_close_n(struct e1inp_line *line, uint8_t trx_nr)
+{
+	struct ipaccess_line *il;
+
+	if (E1INP_SIGN_RSL+trx_nr-1 >= NUM_E1_TS) {
+		LOGP(DLINP, LOGL_ERROR,
+		     "cannot close RSL BTS link: trx_nr (%d) out of range\n", trx_nr);
+		return -EINVAL;
+	}
+
+	il = line->driver_data;
+	if (!il)
+		return 0; /* Nothing to do, no lines created */
+
+	if (il->ipa_cli[1 + trx_nr]) {
+		ipa_client_conn_close(il->ipa_cli[1 + trx_nr]);
+		ipa_client_conn_destroy(il->ipa_cli[1 + trx_nr]);
+		il->ipa_cli[1 + trx_nr] = NULL;
+	}
+	return 0;
+}
+
 void e1inp_ipaccess_init(void)
 {
 	tall_ipa_ctx = talloc_named_const(libosmo_abis_ctx, 1, "ipa");
